<?php
/**
 * NOTICE OF LICENSE
 *
 * @author    INVERTUS, UAB www.invertus.eu <support@invertus.eu>
 * @copyright Copyright (c) permanent, INVERTUS, UAB
 * @license   Addons PrestaShop license limitation
 * @see       /LICENSE
 *
 * International Registered Trademark & Property of INVERTUS, UAB
 */

namespace Invertus\SwedBank\Service;

use Cart;
use Currency;
use Customer;
use Invertus\SwedBank\Adapter\Configuration;
use Invertus\SwedBank\Adapter\Link;
use Invertus\SwedBank\Adapter\Order;
use Invertus\SwedBank\Adapter\Tools;
use Invertus\SwedBank\Adapter\Validate;
use Invertus\SwedBank\Config\Config;
use SwedBank;
use SwedbankOrderPayment;
use SwedbankPaymentPortal\SharedEntity\Authentication;

class OrderHandler
{
    /** @var  Configuration */
    private $configuration;
    /** @var  Order */
    private $orderAdapter;
    /** @var SwedBank  */
    private $module;
    /** @var Link  */
    private $link;
    /** @var Tools  */
    private $tools;
    /** @var Validate  */
    private $validate;

    public function __construct(
        SwedBank $module,
        Configuration $configuration,
        Order $orderAdapter,
        Link $link,
        Tools $tools,
        Validate $validate
    ) {
        $this->configuration = $configuration;
        $this->orderAdapter = $orderAdapter;
        $this->module = $module;
        $this->link = $link;
        $this->tools = $tools;
        $this->validate = $validate;
    }

    public function processOrderCreate(Cart $cart, Currency $currency, $paymentName)
    {
        $environment = $this->tools->strtoupper($this->module->getEnvironment());
        $customer = new Customer($cart->id_customer);
        $total    = (float) $cart->getOrderTotal();
        $idCart   = $cart->id;
        $idOrderState = (int) $this->configuration->get('SB_ORDER_STATUS_AWAITING_'.$environment);

        $this->module->validateOrder(
            $idCart,
            $idOrderState,
            $total,
            $paymentName,
            null,
            array(),
            $currency->id,
            false,
            $customer->secure_key
        );

        $idOrder = (int) $this->orderAdapter->getOrderByCartId($idCart);
        return $idOrder;
    }

    public function getOrderConfirmationUrl(\Order $order, $idLang, array $extraParams = array())
    {
        return $this->link->getPageLink(
            'order-confirmation',
            true,
            $idLang,
            array_merge(
                array(
                'id_cart' => $order->id_cart,
                'id_module' => $this->module->id,
                'id_order' => $order->id,
                'key' => $order->getCustomer()->secure_key,
                ),
                $extraParams
            )
        );
    }

    public function getOrderDetailLink(Customer $customer, \Order $order, $idLang, array $extraParams = array())
    {
        if ($customer->isGuest()) {
            return $this->getOrderConfirmationUrl($order, $idLang, $extraParams);
        }

        return $this->link->getPageLink(
            'order-detail',
            true,
            $idLang,
            [
                'id_order' => $order->id
            ]
        );
    }

    public function getAuthentication($environment)
    {
        switch ($environment) {
            case SwedBank::ENV_TEST:
                $clientId = Config::CLIENT_ID_TEST;
                $clientPassword = Config::CLIENT_PASSWORD_TEST;
                break;
            default:
                $clientId = Config::CLIENT_ID_LIVE;
                $clientPassword = Config::CLIENT_PASSWORD_LIVE;
        }

        return new Authentication(
            $this->configuration->get($clientId),
            $this->configuration->get($clientPassword)
        );
    }

    /**
     * get order amount. When more currencies will be available, use currency transformation here.
     * @param \Order $order
     * @return float
     */
    public function getOrderAmount(\Order $order)
    {
        return (float) $order->total_paid_tax_incl;
    }

    public function checkCustomerAccessByOrder(\Order $order, \Customer $customer)
    {
        return $order->id_customer == $customer->id ||
            $order->isAssociatedAtGuest($this->tools->getValue('guest_email'));
    }

    public function validatePrice($price)
    {
        if (!is_numeric($price)) {
            return false;
        }

        if ($price < 0) {
            return false;
        }
        return true;
    }

    public function getPaymentLanguage($idLang, $idShop, $type)
    {
        $env = $this->module->getEnvironment();
        $idPaymentLanguage = (int) \SwedbankLanguage::getId($idLang, $idShop, $type, $env);
        $paymentLanguage = new \SwedbankLanguage($idPaymentLanguage);

        if (!$this->validate->isLoadedObject($paymentLanguage)) {
            return false;
        }
        return $paymentLanguage->selected_language;
    }

    public function formatMerchantReference(\Order $order, $idShop)
    {
        $idUnique = $this->tools->passwdGen(4);
        SwedbankOrderPayment::associateUniqueId($order->id, $idShop, $idUnique);
        return sprintf(
            '%s-%s',
            $this->tools->strtoupper($order->reference),
            $this->tools->strtoupper($idUnique)
        );
    }

    public function getMerchantReference($order, $idUnique)
    {
        return sprintf(
            '%s-%s',
            $this->tools->strtoupper($order->reference),
            $this->tools->strtoupper($idUnique)
        );
    }
}
